#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from copy import deepcopy

# Import Pyomo libraries
from pyomo.environ import (
    Set,
    Var,
    check_optimal_termination,
    Param,
    Suffix,
    log,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers.get_solver import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.constants import Constants
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.misc import StrEnum
from idaes.core.util.exceptions import InitializationError, ConfigurationError

import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import ControlVolume0DBlock, InitializationMixin
from watertap.costing.unit_models.ion_exchange import cost_ion_exchange

__author__ = "Kurban Sitterley"


"""
REFERENCES

LeVan, M. D., Carta, G., & Yon, C. M. (2019).
Section 16: Adsorption and Ion Exchange.
Perry's Chemical Engineers' Handbook, 9th Edition.

Crittenden, J. C., Trussell, R. R., Hand, D. W., Howe, K. J., & Tchobanoglous, G. (2012).
Chapter 16: Ion Exchange.
MWH's Water Treatment (pp. 1263-1334): John Wiley & Sons, Inc.

DOWEX Ion Exchange Resins Water Conditioning Manual
https://www.lenntech.com/Data-sheets/Dowex-Ion-Exchange-Resins-Water-Conditioning-Manual-L.pdf

Inamuddin, & Luqman, M. (2012).
Ion Exchange Technology I: Theory and Materials.

Vassilis J. Inglezakis and Stavros G. Poulopoulos
Adsorption, Ion Exchange and Catalysis: Design of Operations and Environmental Applications (2006).
doi.org/10.1016/B978-0-444-52783-7.X5000-9

Michaud, C.F. (2013)
Hydrodynamic Design, Part 8: Flow Through Ion Exchange Beds
Water Conditioning & Purification Magazine (WC&P)
https://wcponline.com/2013/08/06/hydrodynamic-design-part-8-flow-ion-exchange-beds/

Clark, R. M. (1987). 
Evaluating the cost and performance of field-scale granular activated carbon systems. 
Environ Sci Technol, 21(6), 573-580. doi:10.1021/es00160a008

Croll, H. C., Adelman, M. J., Chow, S. J., Schwab, K. J., Capelle, R., Oppenheimer, J., & Jacangelo, J. G. (2023). 
Fundamental kinetic constants for breakthrough of per- and polyfluoroalkyl substances at varying empty bed contact times: 
Theoretical analysis and pilot scale demonstration. 
Chemical Engineering Journal, 464. doi:10.1016/j.cej.2023.142587

"""


_log = idaeslog.getLogger(__name__)


class IonExchangeType(StrEnum):
    anion = "anion"
    cation = "cation"


class RegenerantChem(StrEnum):
    HCl = "HCl"
    NaOH = "NaOH"
    H2SO4 = "H2SO4"
    NaCl = "NaCl"
    MeOH = "MeOH"
    single_use = "single_use"


class IsothermType(StrEnum):
    langmuir = "langmuir"
    freundlich = "freundlich"


@declare_process_block_class("IonExchange0D")
class IonExchangeODData(InitializationMixin, UnitModelBlockData):
    """
    Zero order ion exchange model
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False.""",
        ),
    )

    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False.""",
        ),
    )

    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )

    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.none,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.none.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    CONFIG.declare(
        "target_ion",
        ConfigValue(
            default="Ca_2+",
            domain=str,
            description="Designates targeted species for removal",
        ),
    )

    CONFIG.declare(
        "regenerant",
        ConfigValue(
            default=RegenerantChem.NaCl,
            domain=In(RegenerantChem),
            description="Chemical used for regeneration of fixed bed",
        ),
    )

    CONFIG.declare(
        "hazardous_waste",
        ConfigValue(
            default=False,
            domain=bool,
            description="Designates if resin and residuals contain hazardous material",
        ),
    )

    CONFIG.declare(
        "isotherm",
        ConfigValue(
            default=IsothermType.langmuir,
            domain=In(IsothermType),
            description="Designates the isotherm type to use for equilibrium calculations",
        ),
    )

    def build(self):
        super().build()

        comps = self.config.property_package.component_list
        target_ion = self.config.target_ion

        self.target_ion_set = Set(
            initialize=[target_ion]
        )  # create set for future development of multi-component model
        inerts = comps - self.target_ion_set

        if len(self.target_ion_set) > 1:
            raise ConfigurationError(
                f"IonExchange0D can only accept a single target ion but {len(self.target_ion_set)} were provided."
            )
        if self.config.property_package.charge_comp[target_ion].value > 0:
            self.ion_exchange_type = IonExchangeType.cation
        elif self.config.property_package.charge_comp[target_ion].value < 0:
            self.ion_exchange_type = IonExchangeType.anion
        else:
            raise ConfigurationError("Target ion must have non-zero charge.")

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        self.process_flow = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.process_flow.add_state_blocks(has_phase_equilibrium=False)
        self.process_flow.add_material_balances(
            balance_type=self.config.material_balance_type, has_mass_transfer=True
        )
        self.process_flow.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_enthalpy_transfer=False
        )
        self.process_flow.add_isothermal_assumption()
        self.process_flow.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=False,
        )

        prop_in = self.process_flow.properties_in[0]

        self.add_inlet_port(name="inlet", block=self.process_flow)
        self.add_outlet_port(name="outlet", block=self.process_flow)

        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False

        self.regeneration_stream = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of regeneration stream",
            **tmp_dict,
        )

        regen = self.regeneration_stream[0]

        self.add_outlet_port(name="regen", block=self.regeneration_stream)

        for j in inerts:
            self.process_flow.mass_transfer_term[:, "Liq", j].fix(0)
            regen.get_material_flow_terms("Liq", j).fix(0)

        # ==========PARAMETERS==========

        self.underdrain_h = Param(
            initialize=0.5,
            mutable=True,
            units=pyunits.m,
            doc="Underdrain height",  # Perry's
        )

        self.distributor_h = Param(
            initialize=0.5,
            mutable=True,
            units=pyunits.m,
            doc="Distributor height",  # Perry's
        )

        # Particle Peclet number correlation
        # Eq. 4.100 in Inamuddin/Luqman

        self.Pe_p_A = Param(
            initialize=0.05,
            units=pyunits.dimensionless,
            doc="Peclet particle equation A parameter",
        )

        self.Pe_p_exp = Param(
            initialize=0.48,
            units=pyunits.dimensionless,
            doc="Peclet particle equation exponent",
        )

        # Sherwood number as a function of Reynolds and Schmidt number
        # Table 16-9 in Perry's
        # Wilson and Geankoplis, Ind. Eng. Chem. Fundam., 5, 9 (1966)

        self.Sh_A = Param(
            initialize=2.4,
            units=pyunits.dimensionless,
            doc="Sherwood equation A parameter",
        )

        self.Sh_exp_A = Param(
            initialize=0.66,
            units=pyunits.dimensionless,
            doc="Sherwood equation exponent A",
        )

        self.Sh_exp_B = Param(
            initialize=0.34,
            units=pyunits.dimensionless,
            doc="Sherwood equation exponent B",
        )

        self.Sh_exp_C = Param(
            initialize=0.33,
            units=pyunits.dimensionless,
            doc="Sherwood equation exponent C",
        )

        self.number_columns_redund = Param(
            initialize=1,
            mutable=True,
            units=pyunits.dimensionless,
            doc="Number of redundant columns for ion exchange process",
        )

        # Pressure drop (psi/m of resin bed depth) is a function of loading rate (vel_bed) in m/hr
        # p_drop (psi/m) = p_drop_A + p_drop_B * vel_bed + p_drop_C * vel_bed**2
        # Default is for strong-base type I acrylic anion exchanger resin (A-850, Purolite), @20C
        # Data extracted from MWH Chap 16, Figure 16-14 and fit with Excel

        self.p_drop_A = Param(
            initialize=0.609,
            mutable=True,
            units=pyunits.psi / pyunits.m,
            doc="Pressure drop equation intercept",
        )

        self.p_drop_B = Param(
            initialize=0.173,
            mutable=True,
            units=(pyunits.psi * pyunits.hr) / pyunits.m**2,
            doc="Pressure drop equation B",
        )

        self.p_drop_C = Param(
            initialize=8.28e-4,
            mutable=True,
            units=(pyunits.psi * pyunits.hr**2) / pyunits.m**3,
            doc="Pressure drop equation C",
        )

        self.pump_efficiency = Param(
            initialize=0.8,
            mutable=True,
            units=pyunits.dimensionless,
            doc="Pump efficiency",
        )

        self.t_regen = Param(
            initialize=1800,
            mutable=True,
            units=pyunits.s,
            doc="Regeneration time",
        )

        self.service_to_regen_flow_ratio = Param(
            initialize=3,
            mutable=True,
            units=pyunits.dimensionless,
            doc="Ratio of service flow rate to regeneration flow rate",
        )

        # Bed expansion is calculated as a fraction of the bed_depth
        # These coefficients are used to calculate that fraction (bed_expansion_frac) as a function of backwash rate (bw_rate, m/hr)
        # bed_expansion_frac = bed_expansion_A + bed_expansion_B * bw_rate + bed_expansion_C * bw_rate**2
        # Default is for strong-base type I acrylic anion exchanger resin (A-850, Purolite), @20C
        # Data extracted from MWH Chap 16, Figure 16-15 and fit with Excel

        self.bed_expansion_frac_A = Param(
            initialize=-1.23e-2,
            mutable=True,
            units=pyunits.dimensionless,
            doc="Bed expansion fraction eq intercept",
        )

        self.bed_expansion_frac_B = Param(
            initialize=1.02e-1,
            mutable=True,
            units=pyunits.hr / pyunits.m,
            doc="Bed expansion fraction equation B parameter",
        )

        self.bed_expansion_frac_C = Param(
            initialize=-1.35e-3,
            mutable=True,
            units=pyunits.hr**2 / pyunits.m**2,
            doc="Bed expansion fraction equation C parameter",
        )
        # Rinse, Regen, Backwashing params

        self.rinse_bv = Param(
            initialize=5,
            mutable=True,
            doc="Number of bed volumes for rinse step",
        )

        self.bw_rate = Param(
            initialize=5,
            mutable=True,
            units=pyunits.m / pyunits.hour,
            doc="Backwash loading rate [m/hr]",
        )

        self.t_bw = Param(
            initialize=600,
            mutable=True,
            units=pyunits.s,
            doc="Backwash time",
        )

        # ==========VARIABLES==========
        # COMMON TO LANGMUIR + FREUNDLICH

        self.resin_diam = Var(
            initialize=7e-4,
            bounds=(5e-4, 1.5e-3),  # Perry's
            units=pyunits.m,
            doc="Resin bead diameter",
        )

        self.resin_bulk_dens = Var(
            initialize=0.7,
            bounds=(0.65, 0.95),  # Perry's
            units=pyunits.kg / pyunits.L,
            doc="Resin bulk density",
        )

        self.resin_surf_per_vol = Var(
            initialize=3333.33,
            bounds=(0, 1e5),
            units=pyunits.m**-1,
            doc="Resin surface area per volume",
        )

        self.c_norm = Var(
            self.target_ion_set,
            initialize=0.5,
            bounds=(0, 1),
            units=pyunits.dimensionless,
            doc="Dimensionless (relative) concentration [Ct/C0] of target ion",
        )

        self.bed_vol_tot = Var(
            initialize=2,
            units=pyunits.m**3,
            doc="Total bed volume",
        )

        self.bed_depth = Var(
            initialize=1, bounds=(0, 3.0), units=pyunits.m, doc="Bed depth"  # EPA-WBS
        )

        self.bed_porosity = Var(
            initialize=0.5,
            bounds=(0.3, 0.8),
            units=pyunits.dimensionless,
            doc="Bed porosity",
        )

        self.col_height = Var(
            initialize=2,
            bounds=(0.0, 4.5),  # EPA-WBS
            units=pyunits.m,
            doc="Column height",
        )

        self.col_diam = Var(
            initialize=0.5,
            bounds=(0.0, 4.5),  # EPA-WBS
            units=pyunits.m,
            doc="Column diameter",
        )

        self.col_height_to_diam_ratio = Var(
            initialize=1,
            bounds=(1, 100),
            units=pyunits.dimensionless,
            doc="Min ratio of bed depth to diameter",
        )

        self.number_columns = Var(
            initialize=2,
            bounds=(1, None),
            units=pyunits.dimensionless,
            doc="Number of operational columns for ion exchange process",
        )

        self.t_breakthru = Var(
            initialize=1e5,  # DOW, ~7 weeks max breakthru time
            bounds=(0, None),
            units=pyunits.s,
            doc="Breakthrough time",
        )

        self.ebct = Var(
            initialize=520,
            bounds=(90, None),
            units=pyunits.s,
            doc="Empty bed contact time",
        )

        # ====== Hydrodynamic variables ====== #

        self.vel_bed = Var(
            initialize=0.0086,
            bounds=(0, 0.01),  # MWH, Perry's
            units=pyunits.m / pyunits.s,
            doc="Superficial velocity through bed",
        )

        self.service_flow_rate = Var(
            initialize=10,
            bounds=(1, 40),
            units=pyunits.hr**-1,
            doc="Service flow rate [BV/hr]",
        )

        # ====== Dimensionless variables ====== #

        self.N_Re = Var(
            initialize=4.3,
            bounds=(0.001, 60),  # Perry's - bounds relevant to N_Sh regression
            units=pyunits.dimensionless,
            doc="Reynolds number",
        )

        self.N_Sc = Var(
            self.target_ion_set,
            initialize=700,
            units=pyunits.dimensionless,
            doc="Schmidt number",
        )

        self.N_Sh = Var(
            self.target_ion_set,
            initialize=30,
            units=pyunits.dimensionless,
            doc="Sherwood number",
        )

        self.N_Pe_particle = Var(
            initialize=0.1,
            units=pyunits.dimensionless,
            doc="Peclet particle number",
        )

        self.N_Pe_bed = Var(
            initialize=1000,  # Inamuddin/Luqman - N_Pe_bed > ~100 considered to be plug flow
            units=pyunits.dimensionless,
            doc="Peclet bed number",
        )

        if self.config.isotherm == IsothermType.langmuir:

            self.resin_max_capacity = Var(
                initialize=5,
                units=pyunits.mol / pyunits.kg,
                bounds=(0, None),  # Perry's
                doc="Resin max capacity",
            )

            self.resin_eq_capacity = Var(
                initialize=1,
                units=pyunits.mol / pyunits.kg,
                bounds=(0, None),  # Perry's
                doc="Resin equilibrium capacity",
            )

            self.resin_unused_capacity = Var(
                initialize=1,
                units=pyunits.mol / pyunits.kg,
                bounds=(0, None),  # Perry's
                doc="Resin available capacity",
            )

            self.langmuir = Var(
                self.target_ion_set,
                initialize=0.5,  # La < 1 is favorable isotherm
                bounds=(0, 1.1),
                units=pyunits.dimensionless,
                doc="Langmuir isotherm coefficient",
            )

            self.mass_removed = Var(
                self.target_ion_set,
                initialize=1e6,
                bounds=(0, None),
                units=pyunits.mol,
                doc="Sorbed mass of ion",
            )

            self.num_transfer_units = Var(
                initialize=1e6,
                bounds=(0, None),
                units=pyunits.dimensionless,
                doc="Number of transfer units",
            )

            self.dimensionless_time = Var(
                initialize=1,
                units=pyunits.dimensionless,
                doc="Dimensionless time",
            )

            self.partition_ratio = Var(
                initialize=100,
                bounds=(0, None),
                units=pyunits.dimensionless,
                doc="Partition ratio",
            )

            self.fluid_mass_transfer_coeff = Var(
                self.target_ion_set,
                initialize=1e-3,
                bounds=(0, None),
                units=pyunits.m / pyunits.s,
                doc="Fluid mass transfer coefficient",
            )

        if self.config.isotherm == IsothermType.freundlich:

            self.num_traps = 5  # TODO: make CONFIG option
            self.trap_disc = range(self.num_traps + 1)
            self.trap_index = self.trap_disc[1:]

            self.c_trap_min = Param(  # TODO: make CONFIG option
                initialize=0.01,
                mutable=True,
                doc="Minimum relative breakthrough concentration for estimating area under curve",
            )

            # TODO: use trapezoidal approach for langmuir?
            self.c_traps = Var(
                self.trap_disc,
                initialize=0.5,
                bounds=(0, 1),
                units=pyunits.dimensionless,
                doc="Normalized breakthrough concentrations for estimating area under breakthrough curve",
            )

            self.tb_traps = Var(
                self.trap_disc,
                initialize=1e6,
                bounds=(0, None),
                units=pyunits.second,
                doc="Breakthrough times for estimating area under breakthrough curve",
            )

            self.traps = Var(
                self.trap_index,
                initialize=0.01,
                bounds=(0, 1),
                units=pyunits.dimensionless,
                doc="Trapezoid areas for estimating area under breakthrough curve",
            )

            self.c_traps[0].fix(0)
            self.tb_traps[0].fix(0)

            self.c_norm_avg = Var(
                self.target_ion_set,
                initialize=0.25,
                bounds=(0, 2),
                units=pyunits.dimensionless,
                doc="Sum of trapezoid areas",
            )

            self.freundlich_n = Var(
                initialize=1.5,
                bounds=(0, None),
                units=pyunits.dimensionless,
                doc="Freundlich isotherm exponent",
            )

            self.mass_transfer_coeff = Var(  # k_T
                initialize=0.001,
                units=pyunits.s**-1,
                bounds=(0, None),
                doc="Mass transfer coefficient for Clark model (kT)",
            )

            self.bv = Var(  # BV
                initialize=1e5,
                bounds=(0, None),
                units=pyunits.dimensionless,
                doc="Bed volumes of feed at breakthru concentration",
            )

            self.bv_50 = Var(  # BV_50
                initialize=2e5,
                bounds=(0, None),
                units=pyunits.dimensionless,
                doc="Bed volumes of feed at 50 percent breakthrough",
            )

        # ==========EXPRESSIONS==========

        @self.Expression(doc="Pressure drop")
        def pressure_drop(b):
            vel_bed = pyunits.convert(b.vel_bed, to_units=pyunits.m / pyunits.hr)
            return (
                b.p_drop_A + b.p_drop_B * vel_bed + b.p_drop_C * vel_bed**2
            ) * b.bed_depth  # for 20C;

        @self.Expression(doc="Total bed volume")
        def bed_vol(b):
            return b.bed_vol_tot / b.number_columns

        @self.Expression(doc="Rinse time")
        def t_rinse(b):
            return b.ebct * b.rinse_bv

        @self.Expression(doc="Waste time")
        def t_waste(b):
            return b.t_regen + b.t_bw + b.t_rinse

        if self.config.regenerant == RegenerantChem.single_use:
            self.t_regen.set_value(0)
            self.service_to_regen_flow_ratio.set_value(0)

        if self.config.regenerant != RegenerantChem.single_use:

            # If resin is not single use, add regeneration

            @self.Expression(doc="Regen pump power")
            def regen_pump_power(b):
                return pyunits.convert(
                    (
                        b.pressure_drop
                        * (
                            prop_in.flow_vol_phase["Liq"]
                            / b.service_to_regen_flow_ratio
                        )
                    )
                    / b.pump_efficiency,
                    to_units=pyunits.kilowatts,
                )

            @self.Expression(doc="Regen tank volume")
            def regen_tank_vol(b):
                return (
                    prop_in.flow_vol_phase["Liq"] / b.service_to_regen_flow_ratio
                ) * b.t_regen

        @self.Expression(doc="Backwashing flow rate")
        def bw_flow(b):
            return (
                pyunits.convert(b.bw_rate, to_units=pyunits.m / pyunits.s)
                * (b.bed_vol / b.bed_depth)
                * b.number_columns
            )

        @self.Expression(doc="Bed expansion fraction from backwashing")
        def bed_expansion_frac(b):
            return (
                b.bed_expansion_frac_A
                + b.bed_expansion_frac_B * b.bw_rate
                + b.bed_expansion_frac_C * b.bw_rate**2
            )  # for 20C

        @self.Expression(doc="Rinse flow rate")
        def rinse_flow(b):
            return b.vel_bed * (b.bed_vol / b.bed_depth) * b.number_columns

        @self.Expression(doc="Cycle time")
        def t_cycle(b):
            return b.t_breakthru + b.t_waste

        @self.Expression(doc="Backwash pump power")
        def bw_pump_power(b):
            return pyunits.convert(
                (b.pressure_drop * b.bw_flow) / b.pump_efficiency,
                to_units=pyunits.kilowatts,
            )

        @self.Expression(doc="Rinse pump power")
        def rinse_pump_power(b):
            return pyunits.convert(
                (b.pressure_drop * b.rinse_flow) / b.pump_efficiency,
                to_units=pyunits.kilowatts,
            )

        @self.Constraint(
            self.target_ion_set, doc="Mass transfer for regeneration stream"
        )
        def eq_mass_transfer_regen(b, j):
            return (
                regen.get_material_flow_terms("Liq", j)
                == -b.process_flow.mass_transfer_term[0, "Liq", j]
            )

        @self.Constraint(
            doc="Isothermal assumption for regen stream",
        )
        def eq_isothermal_regen_stream(b):
            return prop_in.temperature == regen.temperature

        @self.Constraint(
            doc="Isobaric assumption for regen stream",
        )
        def eq_isobaric_regen_stream(b):
            return prop_in.pressure == regen.pressure

        @self.Expression(doc="Bed expansion from backwashing")
        def bed_expansion_h(b):
            return b.bed_expansion_frac * b.bed_depth

        @self.Expression(doc="Main pump power")
        def main_pump_power(b):
            return pyunits.convert(
                (b.pressure_drop * prop_in.flow_vol_phase["Liq"]) / b.pump_efficiency,
                to_units=pyunits.kilowatts,
            )

        @self.Expression(doc="Volume per column")
        def col_vol_per(b):
            return b.col_height * (b.bed_vol / b.bed_depth)

        @self.Expression(doc="Total column volume required")
        def col_vol_tot(b):
            return b.number_columns * b.col_vol_per

        @self.Expression(doc="Contact time")
        def t_contact(b):
            return b.ebct * b.bed_porosity

        @self.Expression(doc="Interstitial velocity")
        def vel_inter(b):
            return b.vel_bed / b.bed_porosity

        if self.config.isotherm == IsothermType.langmuir:

            @self.Expression(
                doc="Bed volumes at breakthrough",
            )
            def bv_calc(b):
                return (b.vel_bed * b.t_breakthru) / b.bed_depth

            @self.Expression(doc="Left hand side of constant pattern sol'n")
            def lh(b):
                return b.num_transfer_units * (b.dimensionless_time - 1)

            @self.Expression(
                self.target_ion_set,
                doc="Separation factor calc",
            )
            def separation_factor(b, j):
                return 1 / b.langmuir[j]

            @self.Expression(self.target_ion_set, doc="Rate coefficient")
            def rate_coeff(b, j):
                return (6 * (1 - b.bed_porosity) * b.fluid_mass_transfer_coeff[j]) / (
                    pyunits.convert(
                        b.resin_bulk_dens, to_units=pyunits.kg / pyunits.m**3
                    )
                    * b.resin_diam
                )

            @self.Expression(self.target_ion_set, doc="Height of transfer unit - HTU")
            def HTU(b, j):
                return b.vel_bed / (
                    pyunits.convert(
                        b.resin_bulk_dens, to_units=pyunits.kg / pyunits.m**3
                    )
                    * b.rate_coeff[j]
                )

        # ==========CONSTRAINTS==========

        # =========== DIMENSIONLESS ===========

        @self.Constraint(doc="Reynolds number")
        def eq_Re(b):  # Eq. 3.358, Inglezakis + Poulopoulos
            return b.N_Re == (b.vel_bed * b.resin_diam) / prop_in.visc_k_phase["Liq"]

        @self.Constraint(self.target_ion_set, doc="Schmidt number")
        def eq_Sc(b, j):  # Eq. 3.359, Inglezakis + Poulopoulos
            return (
                b.N_Sc[j]
                == prop_in.visc_k_phase["Liq"] / prop_in.diffus_phase_comp["Liq", j]
            )

        @self.Constraint(self.target_ion_set, doc="Sherwood number")
        def eq_Sh(b, j):  # Eq. 3.346, Inglezakis + Poulopoulos
            return (
                b.N_Sh[j]
                == b.Sh_A
                * b.bed_porosity**b.Sh_exp_A
                * b.N_Re**b.Sh_exp_B
                * b.N_Sc[j] ** b.Sh_exp_C
            )

        @self.Constraint(doc="Bed Peclet number")
        def eq_Pe_bed(b):
            return b.N_Pe_bed == b.N_Pe_particle * (b.bed_depth / b.resin_diam)

        @self.Constraint(doc="Particle Peclet number")
        def eq_Pe_p(b):  # Eq. 3.313, Inglezakis + Poulopoulos, for downflow
            return b.N_Pe_particle == b.Pe_p_A * b.N_Re**b.Pe_p_exp

        # =========== RESIN & COLUMN ===========

        @self.Constraint(doc="Resin bead surface area per volume")
        def eq_resin_surf_per_vol(b):
            return b.resin_surf_per_vol == (6 * (1 - b.bed_porosity)) / b.resin_diam

        @self.Constraint(doc="Empty bed contact time")
        def eq_ebct(b):
            return b.ebct * b.vel_bed == b.bed_depth

        @self.Constraint(doc="Service flow rate")
        def eq_service_flow_rate(b):
            return b.service_flow_rate * b.bed_vol_tot == pyunits.convert(
                prop_in.flow_vol_phase["Liq"],
                to_units=pyunits.m**3 / pyunits.hr,
            )

        @self.Constraint(doc="Flow through bed constraint")
        def eq_bed_flow(b):
            return (
                b.bed_depth * prop_in.flow_vol_phase["Liq"] == b.bed_vol_tot * b.vel_bed
            )

        @self.Constraint(doc="Column height")
        def eq_col_height(b):
            return (
                b.col_height
                == b.bed_depth + b.distributor_h + b.underdrain_h + b.bed_expansion_h
            )

        @self.Constraint(doc="Bed design")
        def eq_bed_design(b):
            return (
                b.bed_depth * Constants.pi * (b.col_diam / 2) ** 2
            ) * b.number_columns == b.bed_vol_tot

        @self.Constraint(doc="Column height to diameter ratio")
        def eq_col_height_to_diam_ratio(b):
            return b.col_height_to_diam_ratio * b.col_diam == b.col_height

        # =========== ISOTHERM SPECIFIC ===========

        if self.config.isotherm == IsothermType.langmuir:

            @self.Constraint(doc="Resin capacity mass balance")
            def eq_resin_cap_balance(b):
                return (
                    b.resin_max_capacity
                    == b.resin_unused_capacity + b.resin_eq_capacity
                )

            @self.Constraint(
                self.target_ion_set,
                doc="Mass transfer term for target ion",
            )
            def eq_mass_transfer_target_lang(b, j):
                return (
                    b.mass_removed[j]
                    == -b.process_flow.mass_transfer_term[0, "Liq", j] * b.t_breakthru
                )

            @self.Constraint(self.target_ion_set, doc="Fluid mass transfer coefficient")
            def eq_fluid_mass_transfer_coeff(b, j):
                return (
                    b.fluid_mass_transfer_coeff[j] * b.resin_diam
                    == prop_in.diffus_phase_comp["Liq", j] * b.N_Sh[j]
                )

            @self.Constraint(doc="Partition ratio")
            def eq_partition_ratio(b):
                return b.partition_ratio * pyunits.convert(
                    prop_in.conc_equiv_phase_comp["Liq", target_ion],
                    to_units=pyunits.mol / pyunits.L,
                ) == (b.resin_eq_capacity * b.resin_bulk_dens)

            @self.Constraint(
                self.target_ion_set, doc="Removed total mass of ion in equivalents"
            )
            def eq_mass_removed(b, j):
                charge = prop_in.charge_comp[j]
                return b.mass_removed[j] * charge == pyunits.convert(
                    b.resin_eq_capacity * b.resin_bulk_dens * b.bed_vol_tot,
                    to_units=pyunits.mol,
                )

            @self.Constraint(
                self.target_ion_set,
                doc="Langmuir isotherm",
            )
            def eq_langmuir(b, j):  # Eq. 4.12, Inglezakis + Poulopoulos
                return (1 / b.langmuir[j]) * (
                    b.c_norm[j] * (1 - b.resin_eq_capacity / b.resin_max_capacity)
                ) == (b.resin_eq_capacity / b.resin_max_capacity * (1 - b.c_norm[j]))

            @self.Constraint(doc="Dimensionless time")
            def eq_dimensionless_time(
                b,
            ):  # Eqs. 16-120, 16-129, Perry's; Eq. 4.136, Inglezakis + Poulopoulos
                return b.dimensionless_time * b.partition_ratio == (
                    (b.vel_inter * b.t_breakthru * b.bed_porosity) / b.bed_depth
                    - b.bed_porosity
                )

            @self.Constraint(
                self.target_ion_set,
                doc="Number of mass-transfer units for fluid-film controlling diffusion",
            )
            def eq_num_transfer_units(
                b, j
            ):  # External mass transfer, Perry's Table 16-13; Eq. 4.137, Inglezakis + Poulopoulos
                return b.num_transfer_units * b.vel_bed == (
                    b.fluid_mass_transfer_coeff[j] * b.resin_surf_per_vol * b.bed_depth
                )

            @self.Constraint(
                self.target_ion_set,
                doc="Constant pattern solution for Langmuir isotherm",
            )
            def eq_constant_pattern_soln(
                b, j
            ):  # Liquid-film diffusion control, Eq. 4.140, Inglezakis + Poulopoulos
                return (
                    b.num_transfer_units * (b.dimensionless_time - 1)
                    == (log(b.c_norm[j]) - b.langmuir[j] * log(1 - b.c_norm[j]))
                    / (1 - b.langmuir[j])
                    + 1
                )

        if self.config.isotherm == IsothermType.freundlich:

            @self.Expression(self.target_ion_set, doc="Breakthrough concentration")
            def c_breakthru(b, j):
                return b.c_norm[j] * prop_in.conc_mass_phase_comp["Liq", j]

            @self.Constraint(doc="Bed volumes at breakthrough")
            def eq_bv(b):
                return b.t_breakthru * b.vel_bed == b.bv * b.bed_depth

            @self.Constraint(
                self.target_ion_set, doc="Clark equation with fundamental constants"
            )  # Croll et al (2023), Eq.9
            def eq_clark(b, j):
                left_side = (
                    (b.mass_transfer_coeff * b.bed_depth * (b.freundlich_n - 1))
                    / (b.bv_50 * b.vel_bed)
                ) * (b.bv_50 - b.bv)

                right_side = log(
                    ((1 / b.c_norm[j]) ** (b.freundlich_n - 1) - 1)
                    / (2 ** (b.freundlich_n - 1) - 1)
                )
                return left_side - right_side == 0

            @self.Constraint(
                self.target_ion_set,
                self.trap_index,
                doc="Evenly spaced c_norm for trapezoids",
            )
            def eq_c_traps(b, j, k):
                return b.c_traps[k] == b.c_trap_min + (b.trap_disc[k] - 1) * (
                    (b.c_norm[j] - b.c_trap_min) / (b.num_traps - 1)
                )

            @self.Constraint(
                self.trap_index,
                doc="Breakthru time calc for trapezoids",
            )
            def eq_tb_traps(b, k):
                bv_traps = (b.tb_traps[k] * b.vel_bed) / b.bed_depth
                left_side = (
                    (b.mass_transfer_coeff * b.bed_depth * (b.freundlich_n - 1))
                    / (b.bv_50 * b.vel_bed)
                ) * (b.bv_50 - bv_traps)

                right_side = log(
                    ((1 / b.c_traps[k]) ** (b.freundlich_n - 1) - 1)
                    / (2 ** (b.freundlich_n - 1) - 1)
                )
                return left_side - right_side == 0

            @self.Constraint(self.trap_index, doc="Area of trapezoids")
            def eq_traps(b, k):
                return b.traps[k] == (b.tb_traps[k] - b.tb_traps[k - 1]) / b.tb_traps[
                    self.num_traps
                ] * ((b.c_traps[k] + b.c_traps[k - 1]) / 2)

            @self.Constraint(
                self.target_ion_set, doc="Average relative effluent concentration"
            )
            def eq_c_norm_avg(b, j):
                return b.c_norm_avg[j] == sum(b.traps[k] for k in b.trap_index)

            @self.Constraint(
                self.target_ion_set,
                doc="CV mass transfer term",
            )
            def eq_mass_transfer_target_fr(b, j):
                return (1 - b.c_norm_avg[j]) * prop_in.get_material_flow_terms(
                    "Liq", j
                ) == -b.process_flow.mass_transfer_term[0, "Liq", j]

    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        flags = self.process_flow.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=True,
        )
        init_log.info("Initialization Step 1a Complete.")

        # ---------------------------------------------------------------------
        # Initialize other state blocks
        # Set state_args from inlet state
        if state_args is None:
            self.state_args = state_args = {}
            state_dict = self.process_flow.properties_in[
                self.flowsheet().config.time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        state_args_out = deepcopy(state_args)

        for p, j in self.process_flow.properties_out.phase_component_set:
            if j in self.target_ion_set:
                state_args_out["flow_mol_phase_comp"][(p, j)] = (
                    state_args["flow_mol_phase_comp"][(p, j)] * 1e-3
                )

        self.process_flow.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )
        init_log.info("Initialization Step 1b Complete.")

        state_args_regen = deepcopy(state_args)

        self.regeneration_stream.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_regen,
        )

        init_log.info("Initialization Step 1c Complete.")

        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
            if not check_optimal_termination(res):
                init_log.warning(
                    f"Trouble solving unit model {self.name}, trying one more time"
                )
                res = opt.solve(self, tee=slc.tee)

        init_log.info("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Release Inlet state
        self.process_flow.properties_in.release_state(flags, outlvl=outlvl)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        target_ion = self.config.target_ion
        isotherm = self.config.isotherm

        if iscale.get_scaling_factor(self.t_breakthru) is None:
            iscale.set_scaling_factor(self.t_breakthru, 1e-6)

        if iscale.get_scaling_factor(self.N_Re) is None:
            iscale.set_scaling_factor(self.N_Re, 1)

        if iscale.get_scaling_factor(self.N_Sc) is None:
            iscale.set_scaling_factor(self.N_Sc, 1e-2)

        if iscale.get_scaling_factor(self.N_Sh) is None:
            iscale.set_scaling_factor(self.N_Sh, 0.1)

        if iscale.get_scaling_factor(self.N_Pe_particle) is None:
            iscale.set_scaling_factor(self.N_Pe_particle, 1e2)

        if iscale.get_scaling_factor(self.N_Pe_bed) is None:
            iscale.set_scaling_factor(self.N_Pe_bed, 1e-3)

        if iscale.get_scaling_factor(self.number_columns) is None:
            iscale.set_scaling_factor(self.number_columns, 1)

        if iscale.get_scaling_factor(self.resin_diam) is None:
            iscale.set_scaling_factor(self.resin_diam, 1e4)

        if iscale.get_scaling_factor(self.resin_bulk_dens) is None:
            iscale.set_scaling_factor(self.resin_bulk_dens, 10)

        if iscale.get_scaling_factor(self.resin_surf_per_vol) is None:
            iscale.set_scaling_factor(self.resin_surf_per_vol, 1e-3)

        if iscale.get_scaling_factor(self.bed_vol_tot) is None:
            iscale.set_scaling_factor(self.bed_vol_tot, 0.1)

        if iscale.get_scaling_factor(self.bed_depth) is None:
            iscale.set_scaling_factor(self.bed_depth, 1)

        if iscale.get_scaling_factor(self.col_height_to_diam_ratio) is None:
            iscale.set_scaling_factor(self.col_height_to_diam_ratio, 0.1)

        if iscale.get_scaling_factor(self.bed_porosity) is None:
            iscale.set_scaling_factor(self.bed_porosity, 10)

        if iscale.get_scaling_factor(self.col_height) is None:
            iscale.set_scaling_factor(self.col_height, 1)

        if iscale.get_scaling_factor(self.col_diam) is None:
            iscale.set_scaling_factor(self.col_diam, 1)

        if iscale.get_scaling_factor(self.service_flow_rate) is None:
            iscale.set_scaling_factor(self.service_flow_rate, 0.1)

        if iscale.get_scaling_factor(self.c_norm) is None:
            iscale.set_scaling_factor(self.c_norm, 10)

        if iscale.get_scaling_factor(self.ebct) is None:
            iscale.set_scaling_factor(self.ebct, 1e-2)

        if iscale.get_scaling_factor(self.vel_bed) is None:
            iscale.set_scaling_factor(self.vel_bed, 1e3)

        # unique scaling for isotherm type
        if isotherm == IsothermType.langmuir:
            if iscale.get_scaling_factor(self.resin_max_capacity) is None:
                iscale.set_scaling_factor(self.resin_max_capacity, 1)

            if iscale.get_scaling_factor(self.resin_eq_capacity) is None:
                iscale.set_scaling_factor(self.resin_eq_capacity, 1)

            if iscale.get_scaling_factor(self.resin_unused_capacity) is None:
                iscale.set_scaling_factor(self.resin_unused_capacity, 1)

            if iscale.get_scaling_factor(self.langmuir[target_ion]) is None:
                iscale.set_scaling_factor(self.langmuir[target_ion], 10)

            if iscale.get_scaling_factor(self.num_transfer_units) is None:
                iscale.set_scaling_factor(self.num_transfer_units, 1e-3)

            if iscale.get_scaling_factor(self.partition_ratio) is None:
                iscale.set_scaling_factor(self.partition_ratio, 1e-3)

            if iscale.get_scaling_factor(self.fluid_mass_transfer_coeff) is None:
                iscale.set_scaling_factor(self.fluid_mass_transfer_coeff, 1e5)

            if iscale.get_scaling_factor(self.mass_removed) is None:
                iscale.set_scaling_factor(self.mass_removed, 1e-6)

        if isotherm == IsothermType.freundlich:

            if iscale.get_scaling_factor(self.freundlich_n) is None:
                iscale.set_scaling_factor(self.freundlich_n, 0.1)

            if iscale.get_scaling_factor(self.mass_transfer_coeff) is None:
                iscale.set_scaling_factor(self.mass_transfer_coeff, 10)

            if iscale.get_scaling_factor(self.bv_50) is None:
                iscale.set_scaling_factor(self.bv_50, 1e-5)

            if iscale.get_scaling_factor(self.bv) is None:
                iscale.set_scaling_factor(self.bv, 1e-5)

            if iscale.get_scaling_factor(self.tb_traps) is None:
                sf = iscale.get_scaling_factor(self.t_breakthru)
                iscale.set_scaling_factor(self.tb_traps, sf)

            if iscale.get_scaling_factor(self.c_traps) is None:
                iscale.set_scaling_factor(self.c_traps, 1)

            if iscale.get_scaling_factor(self.traps) is None:
                iscale.set_scaling_factor(self.traps, 1e3)

            if iscale.get_scaling_factor(self.c_norm_avg) is None:
                iscale.set_scaling_factor(self.c_norm_avg, 1e2)

        # transforming constraints
        if isotherm == IsothermType.langmuir:
            for ind, c in self.eq_num_transfer_units.items():
                if iscale.get_scaling_factor(c) is None:
                    sf = iscale.get_scaling_factor(self.num_transfer_units)
                    iscale.constraint_scaling_transform(c, sf)

            for _, c in self.eq_partition_ratio.items():
                if iscale.get_scaling_factor(c) is None:
                    sf = iscale.get_scaling_factor(
                        self.process_flow.properties_in[0].conc_mol_phase_comp[
                            "Liq", target_ion
                        ]
                    )
                    iscale.constraint_scaling_transform(c, sf)

            for ind, c in self.eq_fluid_mass_transfer_coeff.items():
                if iscale.get_scaling_factor(c) is None:
                    sf = iscale.get_scaling_factor(self.fluid_mass_transfer_coeff[ind])
                    iscale.constraint_scaling_transform(c, sf)

        if isotherm == IsothermType.freundlich:

            for ind, c in self.eq_clark.items():
                if iscale.get_scaling_factor(c) is None:
                    iscale.constraint_scaling_transform(c, 1e-2)

            for ind, c in self.eq_traps.items():
                if iscale.get_scaling_factor(c) is None:
                    iscale.constraint_scaling_transform(c, 1e2)

    def _get_stream_table_contents(self, time_point=0):

        return create_stream_table_dataframe(
            {
                "Feed Inlet": self.inlet,
                "Liquid Outlet": self.outlet,
                "Regen Outlet": self.regen,
            },
            time_point=time_point,
        )

    def _get_performance_contents(self, time_point=0):

        # TODO: add relevant Params, Expressions, differences for CONFIGs
        target_ion = self.config.target_ion
        var_dict = {}
        var_dict["Breakthrough Time"] = self.t_breakthru
        var_dict["EBCT"] = self.ebct
        var_dict[f"Relative Breakthrough Conc. [{target_ion}]"] = self.c_norm[
            target_ion
        ]
        var_dict["Number Columns"] = self.number_columns
        var_dict["Bed Volume Total"] = self.bed_vol_tot
        var_dict["Bed Depth"] = self.bed_depth
        var_dict["Col. Height to Diam. Ratio"] = self.col_height_to_diam_ratio
        var_dict["Bed Porosity"] = self.bed_porosity
        var_dict["Service Flow Rate [BV/hr]"] = self.service_flow_rate
        var_dict["Bed Velocity"] = self.vel_bed
        var_dict["Resin Particle Diameter"] = self.resin_diam
        var_dict["Resin Bulk Density"] = self.resin_bulk_dens
        var_dict[f"Schmidt Number [{target_ion}]"] = self.N_Sc[target_ion]
        var_dict[f"Sherwood Number [{target_ion}]"] = self.N_Sh[target_ion]
        var_dict["Reynolds Number"] = self.N_Re
        var_dict["Peclet Number (bed)"] = self.N_Pe_bed
        var_dict["Peclet Number (particle)"] = self.N_Pe_particle
        if self.config.isotherm == IsothermType.langmuir:
            var_dict["Total Resin Capacity [eq/L]"] = self.resin_max_capacity
            var_dict["Usable Resin Capacity [eq/L]"] = self.resin_eq_capacity
            var_dict["Number Transfer Units"] = self.num_transfer_units
            var_dict["Total Mass Removed [equivalents]"] = self.mass_removed[target_ion]
            var_dict["Dimensionless Time"] = self.dimensionless_time
            var_dict["Partition Ratio"] = self.partition_ratio
            var_dict[f"Langmuir Coeff. [{target_ion}]"] = self.langmuir[target_ion]
            var_dict[
                f"Fluid Mass Transfer Coeff. [{target_ion}]"
            ] = self.fluid_mass_transfer_coeff[target_ion]
        elif self.config.isotherm == IsothermType.freundlich:
            var_dict[f"BV at Breakthrough"] = self.bv
            var_dict[f"BV at 50% Breakthrough"] = self.bv_50
            var_dict[f"Freundlich n"] = self.freundlich_n
            var_dict[f"Clark Mass Transfer Coeff."] = self.mass_transfer_coeff

        return {"vars": var_dict}

    @property
    def default_costing_method(self):
        return cost_ion_exchange



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import functools

import pyomo.environ as pyo


def register_costing_parameter_block(build_rule, parameter_block_name):
    """
    Decorator to create a global-level parameter block on the costing package.

    This decorator is meant to be utilized on a costing method or function in order
    to register parameters (including flows specific to a unit model) which should
    have global-scope within a costing package. In this way, all instances of a unit
    model can share the same global parameters, such as baseline capital costs.

    The decorator takes a `build_rule`, which is a rule for constructing a Pyomo
    `Block`, and a `parameter_block_name`, which determines the name of the constructed
    block. The decorator then puts the new Block onto the costing package. For a unit
    model costing block named `blk`, this is symmatically equivalent to:

    .. testcode::
        :skipif: True

        blk.costing_package.parameter_block_name = Block(rule=build_rule)

    However, this decorator contains a few additional safeguards.

        1. If the wrapped unit model costing function is called more than once,
           the associated parameter block will only be added one time.

        2. Similarly, if a component named `parameter_block_name` already exists
           on the `costing_package`, this decorator will raise an error if the
           parameter block was built with a different `build_rule`. This prevents
           inadvertent name clashes between different parameter blocks.

        3. The decorator fixes any Pyomo Vars it finds on the constructed parameter
           block.

    Use of this decorator allows for defining unit-level costing parameters in the same
    module as the unit-level costing method. It also ensures that unit-level costing
    parameters are only added to the costing package if and only if that particular
    unit model is costed.

    Args:
        build_rule (callable): A function defining a "build rule" for a Pyomo `Block`.
        parameter_block_name (str): The name for the built Block on the costing package.
    """

    def register_costing_parameter_block_decorator(func):
        @functools.wraps(func)
        def add_costing_parameter_block(blk, *args, **kwargs):
            parameter_block = blk.costing_package.component(parameter_block_name)
            if parameter_block is None:
                parameter_block = pyo.Block(rule=build_rule)
                blk.costing_package.add_component(parameter_block_name, parameter_block)
                # fix the parameters in case the build_rule did not
                parameter_block.fix_all_vars()
            elif parameter_block._rule is None or not hasattr(
                parameter_block._rule, "_fcn"
            ):
                raise RuntimeError(
                    "Use the register_costing_parameter_block decorator for specifying"
                    "costing-package-level parameters"
                )
            elif parameter_block._rule._fcn is not build_rule:
                other_rule = parameter_block._rule._fcn
                raise RuntimeError(
                    "Attempting to add identically named costing parameter blocks with "
                    "different build rules to the costing package "
                    f"{blk.costing_package}. Parameter block named "
                    f"{parameter_block_name} was previously built by function "
                    f"{other_rule.__name__} from module {other_rule.__module__}"
                )
            # else parameter_block was constructed by build_rule previously
            return func(blk, *args, **kwargs)

        return add_costing_parameter_block

    return register_costing_parameter_block_decorator


def make_capital_cost_var(blk):
    """
    Generic function for adding a Var named `capital_cost`
    to the unit model costing block.

    Creates a pyomo Var named `capital_cost` with units
    `blk.costing_package.base_currency`.
    """
    blk.capital_cost = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency,
        doc="Unit capital cost",
    )


def make_fixed_operating_cost_var(blk):
    """
    Generic function for adding a Var named `fixed_operating_cost`
    to the unit model costing block.

    Creates a pyomo Var named `fixed_operating_cost` with units
    `blk.costing_package.base_currency / blk.costing_package.base_period`.
    """
    blk.fixed_operating_cost = pyo.Var(
        initialize=1e5,
        domain=pyo.NonNegativeReals,
        units=blk.costing_package.base_currency / blk.costing_package.base_period,
        doc="Unit fixed operating cost",
    )


def cost_membrane(blk, membrane_cost, factor_membrane_replacement):
    """
    Generic function for costing a membrane. Assumes the unit_model
    has an `area` variable or parameter.

    Args:
        membrane_cost: The cost of the membrane in currency per area
        factor_membrane_replacement: Membrane replacement factor
            (fraction of membrane replaced/year)
    """

    make_capital_cost_var(blk)
    make_fixed_operating_cost_var(blk)
    blk.membrane_cost = pyo.Expression(expr=membrane_cost)
    blk.factor_membrane_replacement = pyo.Expression(expr=factor_membrane_replacement)

    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.membrane_cost * blk.unit_model.area,
            to_units=blk.costing_package.base_currency,
        )
    )
    blk.fixed_operating_cost_constraint = pyo.Constraint(
        expr=blk.fixed_operating_cost
        == pyo.units.convert(
            blk.factor_membrane_replacement * blk.membrane_cost * blk.unit_model.area,
            to_units=blk.costing_package.base_currency
            / blk.costing_package.base_period,
        )
    )


def cost_rectifier(blk, power=100 * pyo.units.kW, ac_dc_conversion_efficiency=0.90):
    """
    Method to cost rectifiers for electrified process units that require direct current which must be converted
    from an alternating current source. Note that this should be used solely for units that require the conversion,
    and should not be used universally for electricity requirements.
    Assumes the unit_model has a `power` variable or parameter.

    Args:
        ac_dc_conversion_efficiency: Efficiency of the conversion from AC to DC current
    """

    # create variables on cost block
    make_capital_cost_var(blk)
    blk.ac_dc_conversion_efficiency = pyo.Expression(
        expr=ac_dc_conversion_efficiency,
        doc="fixing unit model vairable for upscaling required power considering "
        "the efficiency of converting alternating to direct current",
    )
    blk.ac_power = pyo.Var(
        initialize=100,
        domain=pyo.NonNegativeReals,
        units=pyo.units.kW,
        doc="Unit AC power",
    )

    # use unit.power variable in conversion with efficiency
    blk.power_conversion = pyo.Constraint(
        expr=blk.ac_power * blk.ac_dc_conversion_efficiency
        == pyo.units.convert(power, to_units=pyo.units.kW)
    )

    # USD_2021 embedded in equation
    rectifier_cost_coeff = {0: 508.6, 1: 2810}
    blk.rectifier_cost_coeff = pyo.Var(
        rectifier_cost_coeff.keys(),
        initialize=rectifier_cost_coeff,
        units=pyo.units.dimensionless,
        doc="Rectifier cost coefficients",
    )
    blk.capital_cost_rectifier = pyo.Var(
        initialize=100, units=blk.costing_package.base_currency
    )

    # refix variables to appropriate costing parameters
    for index, var in blk.rectifier_cost_coeff.items():
        var.fix(rectifier_cost_coeff[index])

    # calculate capital cost
    blk.capital_cost_rectifier_constraint = pyo.Constraint(
        expr=blk.capital_cost_rectifier
        == pyo.units.convert(
            pyo.units.USD_2021
            * (
                blk.rectifier_cost_coeff[1]
                + (blk.rectifier_cost_coeff[0] * (blk.ac_power * pyo.units.kW**-1))
            ),
            to_units=blk.costing_package.base_currency,
        )
    )

    # cost electicity flow
    blk.costing_package.cost_flow(blk.ac_power, "electricity")


def cost_by_flow_volume(blk, flow_cost, flow_to_cost):
    """
    Generic function for costing by flow volume.

    Args:
        flow_cost: The cost of the device in [currency]/([volume]/[time])
        flow_to_cost: The flow costed in [volume]/[time]
    """
    make_capital_cost_var(blk)
    blk.flow_cost = pyo.Expression(expr=flow_cost)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.flow_cost * flow_to_cost, to_units=blk.costing_package.base_currency
        )
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
import watertap.property_models.seawater_prop_pack as props
from idaes.models.properties.tests.test_harness import (
    PropertyTestHarness as PropertyTestHarness_idaes,
)
from watertap.property_models.tests.property_test_harness import (
    PropertyTestHarness,
    PropertyRegressionTest,
    PropertyCalculateStateTest,
)

# -----------------------------------------------------------------------------
class TestSeawaterProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestSeawaterProperty(PropertyTestHarness):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}
        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e2,
        }
        self.stateblock_statistics = {
            "number_variables": 26,
            "number_total_constraints": 22,
            "number_unused_variables": 0,
            "default_degrees_of_freedom": 4,
        }  # 4 state vars, but pressure is not active
        self.default_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.965,
            ("mass_frac_phase_comp", ("Liq", "TDS")): 0.035,
            ("dens_mass_phase", "Liq"): 1023.5,
            ("dens_mass_solvent", None): 996.9,
            ("flow_vol_phase", "Liq"): 9.770e-4,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 987.7,
            ("conc_mass_phase_comp", ("Liq", "TDS")): 35.82,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 53.57,
            ("flow_mol_phase_comp", ("Liq", "TDS")): 1.1145,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9796,
            ("mole_frac_phase_comp", ("Liq", "TDS")): 2.038e-2,
            ("molality_phase_comp", ("Liq", "TDS")): 1.155,
            ("visc_d_phase", "Liq"): 9.588e-4,
            ("osm_coeff", None): 0.9068,
            ("pressure_osm_phase", "Liq"): 2.588e6,
            ("enth_mass_phase", "Liq"): 9.896e4,
            ("pressure_sat", None): 3111,
            ("cp_mass_phase", "Liq"): 4001,
            ("therm_cond_phase", "Liq"): 0.6086,
            ("dh_vap_mass", None): 2.356e6,
            ("diffus_phase_comp", ("Liq", "TDS")): 1.471e-9,
            ("boiling_point_elevation_phase", "Liq"): 0.3093,
        }


@pytest.mark.component
class TestSeawaterPropertySolution_1(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e2,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.99,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 0.01,
            ("temperature", None): 273.15 + 50,
            ("pressure", None): 2e5,
        }
        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.99,
            ("mass_frac_phase_comp", ("Liq", "TDS")): 0.01,
            ("dens_mass_phase", "Liq"): 995.4,
            ("dens_mass_solvent", None): 988.0,
            ("flow_vol_phase", "Liq"): 1.005e-3,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 985.5,
            ("conc_mass_phase_comp", ("Liq", "TDS")): 9.954,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 54.95,
            ("flow_mol_phase_comp", ("Liq", "TDS")): 0.3184,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9942,
            ("mole_frac_phase_comp", ("Liq", "TDS")): 5.761e-3,
            ("molality_phase_comp", ("Liq", "TDS")): 0.3216,
            ("visc_d_phase", "Liq"): 5.596e-4,
            ("osm_coeff", None): 0.9029,
            ("pressure_osm_phase", "Liq"): 7.710e5,
            ("enth_mass_phase", "Liq"): 2.057e5,
            ("pressure_sat", None): 1.229e4,
            ("cp_mass_phase", "Liq"): 4.130e3,
            ("therm_cond_phase", "Liq"): 0.6400,
            ("dh_vap_mass", None): 2.358e6,
            ("diffus_phase_comp", ("Liq", "TDS")): 1.493e-9,
            ("boiling_point_elevation_phase", "Liq"): 0.0989,
        }


@pytest.mark.component
class TestSeawaterPropertySolution_2(PropertyRegressionTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e2,
        }
        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.95,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 0.05,
            ("temperature", None): 273.15 + 10,
            ("pressure", None): 100e5,
        }
        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.95,
            ("mass_frac_phase_comp", ("Liq", "TDS")): 0.05,
            ("dens_mass_phase", "Liq"): 1.039e3,
            ("dens_mass_solvent", None): 999.5,
            ("flow_vol_phase", "Liq"): 9.628e-4,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 986.8,
            ("conc_mass_phase_comp", ("Liq", "TDS")): 51.93,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 52.73,
            ("flow_mol_phase_comp", ("Liq", "TDS")): 1.592,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9707,
            ("mole_frac_phase_comp", ("Liq", "TDS")): 2.931e-2,
            ("molality_phase_comp", ("Liq", "TDS")): 1.676,
            ("visc_d_phase", "Liq"): 1.443e-3,
            ("osm_coeff", None): 0.9106,
            ("pressure_osm_phase", "Liq"): 3.591e6,
            ("enth_mass_phase", "Liq"): 4.783e4,
            ("pressure_sat", None): 1.194e3,
            ("cp_mass_phase", "Liq"): 3.916e3,
            ("therm_cond_phase", "Liq"): 0.5854,
            ("dh_vap_mass", None): 2.353e6,
            ("diffus_phase_comp", ("Liq", "TDS")): 1.471e-9,
            ("boiling_point_elevation_phase", "Liq"): 0.4069,
        }


# @pytest.mark.component
# class TestSeawaterPropertySolution_3(PropertyRegressionTest):
#     def configure(self):
#         self.prop_pack = props.SeawaterParameterBlock
#         self.param_args = {}
#
#         self.solver = "ipopt"
#         self.optarg = {"nlp_scaling_method": "user-scaling"}
#
#         self.scaling_args = {
#             ("flow_vol_phase", "Liq"): 1,
#         }
#         self.state_args = {
#             ("flow_vol_phase", "Liq"): 10,
#             ("mass_frac_phase_comp", ("Liq", "TDS")): 0.04,
#             ("temperature", None): 273.15 + 10,
#             ("pressure", None): 2e6,
#         }
#         self.regression_solution = {
#             ("mass_frac_phase_comp", ("Liq", "H2O")): 0.96,
#             ("mass_frac_phase_comp", ("Liq", "TDS")): 0.04,
#             ("dens_mass_phase", "Liq"): 1.039e3,
#             ("dens_mass_solvent", None): 999.5,
#             ("flow_vol_phase", "Liq"): 9.628e-4,
#             ("conc_mass_phase_comp", ("Liq", "H2O")): 986.8,
#             ("conc_mass_phase_comp", ("Liq", "TDS")): 51.93,
#             ("flow_mol_phase_comp", ("Liq", "H2O")): 52.73,
#             ("flow_mol_phase_comp", ("Liq", "TDS")): 1.592,
#             ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9707,
#             ("mole_frac_phase_comp", ("Liq", "TDS")): 2.931e-2,
#             ("molality_phase_comp", ("Liq", "TDS")): 1.676,
#             ("visc_d_phase", "Liq"): 1.443e-3,
#             ("osm_coeff", None): 0.9106,
#             ("pressure_osm_phase", "Liq"): 3.591e6,
#             ("enth_mass_phase", "Liq"): 4.783e4,
#             ("pressure_sat", None): 1.194e3,
#             ("cp_mass_phase", "Liq"): 3.916e3,
#             ("therm_cond_phase", "Liq"): 0.5854,
#             ("dh_vap_mass", None): 2.353e6,
#             ("diffus_phase_comp", ("Liq", "TDS")): 1.471e-9,
#             ("boiling_point_elevation_phase", "Liq"): 0.4069,
#         }


@pytest.mark.component
class TestSeawaterCalculateState_1(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e-1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e1,
        }
        self.var_args = {
            ("flow_vol_phase", "Liq"): 2e-2,
            ("mass_frac_phase_comp", ("Liq", "TDS")): 0.05,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }
        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 19.66,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1.035,
        }


@pytest.mark.component
class TestNaClCalculateState_2(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e2,
        }
        self.var_args = {
            ("flow_vol_phase", "Liq"): 1e-3,
            ("pressure_osm_phase", "Liq"): 100e5,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }
        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.9604,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 0.1231,
        }


@pytest.mark.component
class TestNaClCalculateState_3(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e2,
        }
        self.var_args = {
            ("flow_vol_phase", "Liq"): 1e-3,
            ("mass_frac_phase_comp", ("Liq", "TDS")): 0.06,
            ("pressure_sat", None): 3e4,
            ("pressure", None): 5e5,
        }
        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.9605,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 0.0613,
            ("temperature", None): 343.05,
        }


@pytest.mark.component
class TestSeawaterCalculateState_4(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e-4,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e1,
            ("enth_mass_phase", ("Liq", "TDS")): 1e-4,
        }
        self.var_args = {
            ("flow_vol_phase", "Liq"): 10,
            ("mass_frac_phase_comp", ("Liq", "TDS")): 1e-5,
            ("temperature", None): 273.15 + 10,
            ("pressure", None): 2e6,
        }
        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 9995.07,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 0.09995,
            ("enth_mass_phase", "Liq"): 4.3945e4,
        }


@pytest.mark.component
class TestSeawaterCalculateState_5(PropertyCalculateStateTest):
    def configure(self):
        self.prop_pack = props.SeawaterParameterBlock
        self.param_args = {}

        self.solver = "ipopt"
        self.optarg = {"nlp_scaling_method": "user-scaling"}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e-3,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1e-3,
            ("enth_mass_phase", ("Liq", "TDS")): 1e-5,
        }
        self.var_args = {
            ("flow_vol_phase", "Liq"): 10,
            ("mass_frac_phase_comp", ("Liq", "TDS")): 0.12,
            ("temperature", None): 273.15 + 120,
            ("pressure", None): 1.2e7,
        }
        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 9091.08,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 1239.69,
            ("enth_mass_phase", "Liq"): 3.8562e5,
        }



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    cost_by_flow_volume,
)


def build_cstr_cost_param_block(blk):
    # Source: https://www.fwrj.com/articles/9812.pdf
    blk.sizing_cost = pyo.Var(
        initialize=0.34,
        doc="Reactor sizing cost",
        units=pyo.units.USD_1998 / pyo.units.m**3,
    )


@register_costing_parameter_block(
    build_rule=build_cstr_cost_param_block,
    parameter_block_name="cstr",
)
def cost_cstr(blk):
    """
    CSTR costing method
    """
    cost_by_flow_volume(
        blk,
        blk.unit_model.hydraulic_retention_time[0]
        * blk.costing_package.cstr.sizing_cost,
        pyo.units.convert(
            blk.unit_model.control_volume.properties_in[0].flow_vol,
            (pyo.units.meter**3 / pyo.units.hours),
        ),
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from idaes.core.solvers import get_solver
from pyomo.environ import value, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.swine_wwt.swine_wwt import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
    display_metrics_results,
    display_additional_results,
)

solver = get_solver()


class Test_Swine_WWT_Flowsheet:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build()
        return m  # return model

    @pytest.mark.unit()
    def test_build(self, system_frame):
        m = system_frame
        assert_units_consistent(m)
        assert_degrees_of_freedom(m, 53)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        assert_degrees_of_freedom(m, 0)
        initialize_system(m)

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame
        results = solve(m)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        assert_units_consistent(m)
        m.fs.costing.initialize()
        results = solve(m)

        assert_optimal_termination(results)
        # check costing
        lev_costs = {
            "LCOW": 105.5838,
            "LCOH2": 205.8723,
            "LCON": 318.6941,
            "LCOVFA": 36.9335,
            "LCOP": 707.4058,
            "LCOCOD": 13.3432,
            "LCOT": 33.3369,
        }
        for k, v in lev_costs.items():
            assert value(getattr(m.fs.costing.levelized_costs, k)) == pytest.approx(
                v, rel=1e-3
            )

    @pytest.mark.component
    def test_display(self, system_frame):
        m = system_frame
        display_metrics_results(m)
        display_additional_results(m)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Utility functions for EDB tests
"""
import pytest
from pyomo.common.dependencies import attempt_import

from watertap.edb.db_api import ElectrolyteDB

mongomock, mongomock_available = attempt_import("mongomock")


class MockDB(ElectrolyteDB):
    def __init__(self, db="foo", **kwargs):
        if not mongomock_available:
            pytest.skip(reason="mongomock (EDB optional dependency) not available")

        self._client = mongomock.MongoClient()
        self._db = getattr(self._client, db)
        # note: don't call superclass!
        self._database_name = db
        self._server_url = "mock"


@pytest.fixture
def mockdb():
    return MockDB()


def dict_diff(d1, d2, result=[], pfx=""):
    if isinstance(d1, list) and isinstance(d2, list):
        if len(d1) != len(d2):
            result.append(
                f"Array length at {pfx} first({len(d1)}) != second({len(d2)})"
            )
        else:
            pass  # good enough
    elif not isinstance(d1, dict) or not isinstance(d2, dict):
        if type(d1) == type(d2):
            same = None
            try:
                same = d1 == d2
            except:  # cannot compare them
                pass  # good enough
            if same is False:
                result.append(f"value at {pfx} first != second")
        else:
            result.append(
                f"type of value at {pfx} first({type(d1)} != second({type(d2)}"
            )
    else:
        if set(d1.keys()) != set(d2.keys()):
            for k in d1:
                if k not in d2:
                    result.append(f"{pfx}{k} in first, not in second")
            for k in d2:
                if k not in d1:
                    result.append(f"{pfx}{k} in first, not in second")
        for k in d1:
            if k in d2:
                pfx += f"{k}."
                dict_diff(d1[k], d2[k], result=result, pfx=pfx)
    return result



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a water pumping station.
"""

from pyomo.environ import units as pyunits, Var
from pyomo.common.config import ConfigValue, In
from idaes.core import declare_process_block_class
from watertap.core import build_pt, ZeroOrderBaseData
from watertap.core.zero_order_electricity import _common

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("WaterPumpingStationZO")
class WaterPumpingStationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for SW onshore intake operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()
    CONFIG.declare(
        "fix_pump_power",
        ConfigValue(
            default=True,
            domain=In([True, False]),
            description="Boolean flag for fixing pump power directly.",
            doc="""Indicates whether pump power should be fixed by the user or not.
        **default** - True.
        **Valid values:** {
        **True** -  pump power (variable name "electricity") will be fixed by user and lift_height will not be fixed,
        **False** - pump power (variable name "electricity") is left unfixed and lift_height will be fixed,}""",
        ),
    )

    def build(self):
        super().build()

        self._tech_type = "water_pumping_station"

        build_pt(self)

        # create electricity variable and add to performance dictionary
        _common(self)

        self.lift_height = Var(
            self.flowsheet().time,
            initialize=100,
            units=pyunits.feet,
            doc="Lift height for pump",
        )
        self.eta_pump = Var(
            self.flowsheet().time,
            initialize=0.9,
            units=pyunits.dimensionless,
            doc="Efficiency of pump",
        )
        self.eta_motor = Var(
            self.flowsheet().time,
            initialize=0.9,
            units=pyunits.dimensionless,
            doc="Efficiency of motor",
        )

        self._fixed_perf_vars.append(self.eta_pump)
        self._fixed_perf_vars.append(self.eta_motor)

        if not self.config.fix_pump_power:
            self._fixed_perf_vars.append(self.lift_height)
        else:
            self._fixed_perf_vars.append(self.electricity)

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on " "pump flowrate.",
        )
        def electricity_consumption(b, t):
            A = (
                3960
                * pyunits.gallon
                * pyunits.foot
                / pyunits.minute
                / pyunits.horsepower
            )
            return b.electricity[t] == pyunits.convert(
                b.properties[t].flow_vol
                * b.lift_height[t]
                / (A * b.eta_pump[t] * b.eta_motor[t]),
                to_units=pyunits.kW,
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.tools.parameter_sweep import (
    LinearSample,
    parameter_sweep,
)
import watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.suboxic_activated_sludge_process.suboxic_activated_sludge_process as suboxic_activated_sludge_process


def set_up_sensitivity(m):
    outputs = {}
    optimize_kwargs = {"fail_flag": False}
    opt_function = suboxic_activated_sludge_process.solve

    # create outputs
    outputs["LCOW"] = m.fs.costing.LCOW

    return outputs, optimize_kwargs, opt_function


def run_analysis(case_num=1, nx=11, interpolate_nan_outputs=True):

    m = suboxic_activated_sludge_process.main()[0]

    outputs, optimize_kwargs, opt_function = set_up_sensitivity(m)

    sweep_params = {}
    if case_num == 1:
        m.fs.suboxicASM.removal_frac_mass_comp[0, "bod"].unfix()
        sweep_params["bod_removal"] = LinearSample(
            m.fs.suboxicASM.removal_frac_mass_comp[0, "bod"], 0.9363, 0.9904, nx
        )
    elif case_num == 2:
        m.fs.suboxicASM.removal_frac_mass_comp[0, "tss"].unfix()
        sweep_params["tss_removal"] = LinearSample(
            m.fs.suboxicASM.removal_frac_mass_comp[0, "tss"], 0.9405, 0.9940, nx
        )
    elif case_num == 3:
        m.fs.suboxicASM.removal_frac_mass_comp[0, "tkn"].unfix()
        sweep_params["tkn_removal"] = LinearSample(
            m.fs.suboxicASM.removal_frac_mass_comp[0, "tkn"], 0.8846, 0.9615, nx
        )
    elif case_num == 4:
        m.fs.costing.electricity_cost.unfix()
        sweep_params["electricity_cost"] = LinearSample(
            m.fs.costing.electricity_cost, 0.05, 0.12, nx
        )
    else:
        raise ValueError("case_num = %d not recognized." % (case_num))

    output_filename = "sensitivity_" + str(case_num) + ".csv"

    global_results = parameter_sweep(
        m,
        sweep_params,
        outputs,
        csv_results_file_name=output_filename,
        optimize_function=opt_function,
        optimize_kwargs=optimize_kwargs,
        interpolate_nan_outputs=interpolate_nan_outputs,
    )

    return global_results, sweep_params, m


if __name__ == "__main__":
    results, sweep_params, m = run_analysis()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    TerminationCondition,
    SolverStatus,
    value,
    Var,
)
from pyomo.network import Port
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from watertap.unit_models.nanofiltration_0D import NanoFiltration0D
import watertap.property_models.NaCl_prop_pack as props

from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    unscaled_variables_generator,
    unscaled_constraints_generator,
    badly_scaled_var_generator,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestNanoFiltration:
    @pytest.fixture(scope="class")
    def NF_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = NanoFiltration0D(
            property_package=m.fs.properties, has_pressure_change=True
        )

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_NaCl = 0.035
        feed_pressure = 6e5
        feed_temperature = 273.15 + 25
        membrane_pressure_drop = 1e5
        membrane_area = 50 * feed_flow_mass
        A = 3.77e-11
        B = 4.724e-5
        sigma = 0.28
        pressure_atmospheric = 101325

        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
            feed_flow_mass * feed_mass_frac_NaCl
        )
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
            feed_flow_mass * feed_mass_frac_H2O
        )
        m.fs.unit.inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.deltaP.fix(-membrane_pressure_drop)
        m.fs.unit.area.fix(membrane_area)
        m.fs.unit.A_comp.fix(A)
        m.fs.unit.B_comp.fix(B)
        m.fs.unit.sigma.fix(sigma)
        m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
        return m

    @pytest.mark.unit
    def test_config(self, NF_frame):
        m = NF_frame
        # check unit config arguments
        assert len(m.fs.unit.config) == 8

        assert not m.fs.unit.config.dynamic
        assert not m.fs.unit.config.has_holdup
        assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
        assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
        assert (
            m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
        )
        assert m.fs.unit.config.has_pressure_change
        assert m.fs.unit.config.property_package is m.fs.properties

    @pytest.mark.unit
    def test_build(self, NF_frame):
        m = NF_frame

        # test ports and variables
        port_lst = ["inlet", "retentate", "permeate"]
        port_vars_lst = ["flow_mass_phase_comp", "pressure", "temperature"]
        for port_str in port_lst:
            assert hasattr(m.fs.unit, port_str)
            port = getattr(m.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)
            for var_str in port_vars_lst:
                assert hasattr(port, var_str)
                var = getattr(port, var_str)
                assert isinstance(var, Var)

        # test unit objects (including parameters, variables, and constraints)
        unit_objs_lst = [
            "A_comp",
            "B_comp",
            "sigma",
            "dens_solvent",
            "flux_mass_phase_comp_in",
            "flux_mass_phase_comp_out",
            "avg_conc_mass_phase_comp_in",
            "avg_conc_mass_phase_comp_out",
            "area",
            "deltaP",
            "mass_transfer_phase_comp",
            "flux_mass_phase_comp_avg",
            "eq_mass_transfer_term",
            "eq_permeate_production",
            "eq_flux_in",
            "eq_flux_out",
            "eq_avg_conc_in",
            "eq_avg_conc_out",
            "eq_connect_mass_transfer",
            "eq_connect_enthalpy_transfer",
            "eq_permeate_isothermal",
        ]
        for obj_str in unit_objs_lst:
            assert hasattr(m.fs.unit, obj_str)

        # test state block objects
        cv_name = "feed_side"
        cv_stateblock_lst = ["properties_in", "properties_out"]
        stateblock_objs_lst = [
            "flow_mass_phase_comp",
            "pressure",
            "temperature",
            "pressure_osm_phase",
            "osm_coeff",
            "mass_frac_phase_comp",
            "conc_mass_phase_comp",
            "dens_mass_phase",
            "enth_mass_phase",
            "eq_pressure_osm_phase",
            "eq_osm_coeff",
            "eq_mass_frac_phase_comp",
            "eq_conc_mass_phase_comp",
            "eq_dens_mass_phase",
            "eq_enth_mass_phase",
        ]
        # control volume
        assert hasattr(m.fs.unit, cv_name)
        cv_blk = getattr(m.fs.unit, cv_name)
        for blk_str in cv_stateblock_lst:
            assert hasattr(cv_blk, blk_str)
            blk = getattr(cv_blk, blk_str)
            for obj_str in stateblock_objs_lst:
                assert hasattr(blk[0], obj_str)
        # permeate
        assert hasattr(m.fs.unit, "properties_permeate")
        blk = getattr(m.fs.unit, "properties_permeate")
        for var_str in stateblock_objs_lst:
            assert hasattr(blk[0], var_str)

        # test statistics
        assert number_variables(m) == 68
        assert number_total_constraints(m) == 45
        assert number_unused_variables(m) == 7  # vars from property package parameters

    @pytest.mark.unit
    def test_dof(self, NF_frame):
        m = NF_frame
        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_calculate_scaling(self, NF_frame):
        m = NF_frame
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.component
    def test_initialize(self, NF_frame):
        initialization_tester(NF_frame)

    @pytest.mark.component
    def test_var_scaling(self, NF_frame):
        m = NF_frame
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_solve(self, NF_frame):
        m = NF_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_conservation(self, NF_frame):
        m = NF_frame
        b = m.fs.unit
        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.properties_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-6
        )

        assert (
            abs(
                value(
                    flow_mass_inlet
                    * b.feed_side.properties_in[0].enth_mass_phase["Liq"]
                    - flow_mass_retentate
                    * b.feed_side.properties_out[0].enth_mass_phase["Liq"]
                    - flow_mass_permeate
                    * b.properties_permeate[0].enth_mass_phase["Liq"]
                )
            )
            <= 1e-6
        )

    @pytest.mark.component
    def test_solution(self, NF_frame):
        m = NF_frame
        assert pytest.approx(1.079e-2, rel=1e-3) == value(
            m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]
        )
        assert pytest.approx(3.435e-4, rel=1e-3) == value(
            m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]
        )
        assert pytest.approx(0.5396, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        )
        assert pytest.approx(1.717e-2, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        )

    @pytest.mark.unit
    def test_report(self, NF_frame):
        NF_frame.fs.unit.report()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains the methods for constructing the material balances for
zero-order pass-through unit models (i.e. units with a single inlet and single
outlet where flow and composition do not change, such as pumps).
"""
from types import MethodType
from pyomo.environ import check_optimal_termination

import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import number_activated_constraints
from idaes.core.util.exceptions import InitializationError

# Some more inforation about this module
__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


def build_pt(self):
    """
    Helper method for constructing material balances for zero-order type models
    with pass-through behavior.

    One StateBlock is added with two corresponding Ports:
        * properties

    No additional variables and constraints are created.

    This method also sets private attributes on the unit model with references
    to the appropriate initialization and scaling methods to use and to return
    the inlet volumetric flow rate.
    """
    self._has_recovery_removal = False
    self._initialize = MethodType(initialize_pt, self)
    self._scaling = MethodType(calculate_scaling_factors_pt, self)

    # Create state blocks for inlet and outlet
    tmp_dict = dict(**self.config.property_package_args)
    tmp_dict["has_phase_equilibrium"] = False
    tmp_dict["defined_state"] = True

    self.properties = self.config.property_package.build_state_block(
        self.flowsheet().time, doc="Material properties in unit", **tmp_dict
    )

    # Create Ports
    self.add_port("inlet", self.properties, doc="Inlet port")
    self.add_port("outlet", self.properties, doc="Outlet port")

    self._stream_table_dict = {"Inlet": self.inlet, "Outlet": self.outlet}

    self._get_Q = MethodType(_get_Q_pt, self)


def initialize_pt(
    blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
):
    """
    Initialization routine for pass-through unit models.

    Keyword Arguments:
        state_args : a dict of arguments to be passed to the property
                       package(s) to provide an initial state for
                       initialization (see documentation of the specific
                       property package) (default = {}).
        outlvl : sets output level of initialization routine
        optarg : solver options dictionary object (default=None, use
                 default solver options)
        solver : str indicating which solver to use during
                 initialization (default = None, use default IDAES solver)

    Returns:
        None
    """
    init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
    solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

    solver_obj = get_solver(solver, optarg)

    # Get initial guesses for inlet if none provided
    if state_args is None:
        state_args = {}
        state_dict = blk.properties[blk.flowsheet().time.first()].define_port_members()

        for k in state_dict.keys():
            if state_dict[k].is_indexed():
                state_args[k] = {}
                for m in state_dict[k].keys():
                    state_args[k][m] = state_dict[k][m].value
            else:
                state_args[k] = state_dict[k].value

    # ---------------------------------------------------------------------
    # Initialize control volume block
    flags = blk.properties.initialize(
        outlvl=outlvl,
        optarg=optarg,
        solver=solver,
        state_args=state_args,
        hold_state=True,
    )
    init_log.info_high("Initialization Step 1 Complete.")

    # ---------------------------------------------------------------------
    # Solve unit
    if number_activated_constraints(blk) > 0:
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = solver_obj.solve(blk, tee=slc.tee)

        init_log.info_high(
            "Initialization Step 2 {}.".format(idaeslog.condition(results))
        )

    # ---------------------------------------------------------------------
    # Release Inlet state
    blk.properties.release_state(flags, outlvl)

    init_log.info("Initialization Complete: {}".format(idaeslog.condition(results)))

    if number_activated_constraints(blk) > 0 and not check_optimal_termination(results):
        raise InitializationError(
            f"{blk.name} failed to initialize successfully. Please check "
            f"the output logs for more information."
        )


def calculate_scaling_factors_pt(self):
    pass


def _get_Q_pt(self, t):
    return self.properties[t].flow_vol



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest

from pyomo.environ import ConcreteModel, Param, Var, units

from watertap.core.util.misc import is_constant_up_to_units


@pytest.mark.unit
def test_is_constant_up_to_units():

    # test float
    assert is_constant_up_to_units(42.0)

    # test int
    assert is_constant_up_to_units(42)

    # test unit-only expression
    assert is_constant_up_to_units(42.0 * units.m**2 / units.s)

    m = ConcreteModel()
    m.fixed_param = Param(initialize=42.0)

    # test non-mutable param
    assert is_constant_up_to_units(m.fixed_param)

    # test non-mutable param with unit-only expression
    assert is_constant_up_to_units(m.fixed_param * units.m**2 / units.s)

    m.fixed_param_2 = Param(initialize=6.28)

    # test expression of fixed params with units
    assert is_constant_up_to_units(
        (m.fixed_param**2) * m.fixed_param_2 * units.m**2
    )

    m.mutable_param = Param(initialize=42, mutable=True)

    # test mutable param
    assert not is_constant_up_to_units(m.mutable_param)

    m.mutable_param_units = Param(initialize=42, units=units.m**2 / units.s)
    # test mutable param with units specified
    assert not is_constant_up_to_units(m.mutable_param_units)

    m.fixed_variable = Var(initialize=42)
    m.fixed_variable.fix()
    m.fixed_variable_units = Var(initialize=42, units=units.m**2 / units.s)
    m.fixed_variable_units.fix()

    m.unfixed_variable = Var(initialize=6.28)
    m.unfixed_variable_units = Var(initialize=6.28, units=units.m**2 / units.s)

    # test variables
    assert not is_constant_up_to_units(m.fixed_variable)
    assert not is_constant_up_to_units(m.fixed_variable_units)
    assert not is_constant_up_to_units(m.unfixed_variable)
    assert not is_constant_up_to_units(m.unfixed_variable_units)

    # test combinations
    assert not is_constant_up_to_units(m.fixed_variable * m.fixed_param * units.m**2)
    assert not is_constant_up_to_units(m.fixed_variable * 42.0 * units.m**2)

    assert not is_constant_up_to_units(
        m.unfixed_variable * m.fixed_param * units.m**2
    )
    assert not is_constant_up_to_units(m.unfixed_variable * 42.0 * units.m**2)

    assert not is_constant_up_to_units(
        m.fixed_variable_units * m.fixed_param * units.m**2
    )
    assert not is_constant_up_to_units(m.fixed_variable_units * 42.0 * units.m**2)

    assert not is_constant_up_to_units(
        m.unfixed_variable_units * m.fixed_param * units.m**2
    )
    assert not is_constant_up_to_units(m.unfixed_variable_units * 42.0 * units.m**2)

    assert not is_constant_up_to_units(
        (1 / m.mutable_param_units) * m.fixed_param * units.m**2
    )
    assert not is_constant_up_to_units(
        (1 / m.mutable_param_units) * m.fixed_variable_units * 42.0 * units.m**2
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a fixed bed unit
operation.
"""

from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_siso, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("FixedBedZO")
class FixedBedZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an Ion exchange unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "fixed_bed"

        build_siso(self)
        constant_intensity(self)
        self.recovery_frac_mass_H2O.fix(1)

        # Chemical demands
        self.acetic_acid_dose = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Dosing rate of acetic acid",
        )
        self.phosphoric_acid_dose = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Dosing rate of phosphoric acid",
        )
        self.ferric_chloride_dose = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Dosing rate of ferric chloride",
        )
        self._fixed_perf_vars.append(self.acetic_acid_dose)
        self._fixed_perf_vars.append(self.phosphoric_acid_dose)
        self._fixed_perf_vars.append(self.ferric_chloride_dose)

        self.acetic_acid_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Consumption rate of acetic acid",
        )
        self.phosphoric_acid_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Consumption rate of phosphoric acid",
        )
        self.ferric_chloride_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Consumption rate of ferric chloride",
        )

        self._perf_var_dict["Acetic Acid Demand"] = self.acetic_acid_demand
        self._perf_var_dict["Phosphoric Acid Demand"] = self.phosphoric_acid_demand
        self._perf_var_dict["Ferric Chlorided Demand"] = self.ferric_chloride_demand

        @self.Constraint(self.flowsheet().time, doc="Acetic acid demand constraint")
        def acetic_acid_demand_equation(b, t):
            return b.acetic_acid_demand[t] == pyunits.convert(
                b.acetic_acid_dose * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

        @self.Constraint(self.flowsheet().time, doc="Phosphoric acid demand constraint")
        def phosphoric_acid_demand_equation(b, t):
            return b.phosphoric_acid_demand[t] == pyunits.convert(
                b.phosphoric_acid_dose * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

        @self.Constraint(self.flowsheet().time, doc="Acetic acid demand constraint")
        def ferric_chloride_demand_equation(b, t):
            return b.ferric_chloride_demand[t] == pyunits.convert(
                b.ferric_chloride_dose * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

        # Activated Carbon demand
        self.activated_carbon_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for activated carbon",
        )
        self.activated_carbon_parameter_a = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for activated carbon demand",
        )
        self.activated_carbon_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for activated carbon demand",
        )
        self._fixed_perf_vars.append(self.activated_carbon_parameter_a)
        self._fixed_perf_vars.append(self.activated_carbon_parameter_b)
        self._perf_var_dict["Activated Carbon Demand"] = self.activated_carbon_demand

        @self.Constraint(
            self.flowsheet().time, doc="Activated carbon demand constraint"
        )
        def activated_carbon_demand_equation(b, t):
            return b.activated_carbon_demand[t] == pyunits.convert(
                b.activated_carbon_parameter_a
                * pyunits.convert(
                    b.properties_in[t].flow_vol / (pyunits.m**3 / pyunits.hour),
                    to_units=pyunits.dimensionless,
                )
                ** b.activated_carbon_parameter_b
                * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

        # Sand demand
        self.sand_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for sand",
        )
        self.sand_parameter_a = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for sand demand",
        )
        self.sand_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for sand demand",
        )
        self._fixed_perf_vars.append(self.sand_parameter_a)
        self._fixed_perf_vars.append(self.sand_parameter_b)
        self._perf_var_dict["Sand Demand"] = self.sand_demand

        @self.Constraint(self.flowsheet().time, doc="Sand demand constraint")
        def sand_demand_equation(b, t):
            return b.sand_demand[t] == pyunits.convert(
                b.sand_parameter_a
                * pyunits.convert(
                    b.properties_in[t].flow_vol / (pyunits.m**3 / pyunits.hour),
                    to_units=pyunits.dimensionless,
                )
                ** b.sand_parameter_b
                * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

        # Anthracite demand
        self.anthracite_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for anthracite",
        )
        self.anthracite_parameter_a = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for anthracite demand",
        )
        self.anthracite_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for anthracite demand",
        )
        self._fixed_perf_vars.append(self.anthracite_parameter_a)
        self._fixed_perf_vars.append(self.anthracite_parameter_b)
        self._perf_var_dict["Anthracite Demand"] = self.anthracite_demand

        @self.Constraint(self.flowsheet().time, doc="Anthracite demand constraint")
        def anthracite_demand_equation(b, t):
            return b.anthracite_demand[t] == pyunits.convert(
                b.anthracite_parameter_a
                * pyunits.convert(
                    b.properties_in[t].flow_vol / (pyunits.m**3 / pyunits.hour),
                    to_units=pyunits.dimensionless,
                )
                ** b.anthracite_parameter_b
                * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

        # Cationic polymer demand
        self.cationic_polymer_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for cationic polymer",
        )
        self.cationic_polymer_parameter_a = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for cationic polymer demand",
        )
        self.cationic_polymer_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for cationic polymer demand",
        )
        self._fixed_perf_vars.append(self.cationic_polymer_parameter_a)
        self._fixed_perf_vars.append(self.cationic_polymer_parameter_b)
        self._perf_var_dict["Cationic Polymer Demand"] = self.cationic_polymer_demand

        @self.Constraint(
            self.flowsheet().time, doc="Cationic Polymer demand constraint"
        )
        def cationic_polymer_demand_equation(b, t):
            return b.cationic_polymer_demand[t] == pyunits.convert(
                b.cationic_polymer_parameter_a
                * pyunits.convert(
                    b.properties_in[t].flow_vol / (pyunits.m**3 / pyunits.hour),
                    to_units=pyunits.dimensionless,
                )
                ** b.cationic_polymer_parameter_b
                * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.hr,
            )

    @property
    def default_costing_method(self):
        return self.cost_fixed_bed

    @staticmethod
    def cost_fixed_bed(blk, number_of_parallel_units=1):
        """
        General method for costing fixed bed units. This primarily calls the
        cost_power_law_flow method.
        This method also registers demand for a number of additional material
        flows.
        Args:
            number_of_parallel_units (int, optional) - cost this unit as
                        number_of_parallel_units parallel units (default: 1)
        """
        t0 = blk.flowsheet().time.first()

        blk.unit_model.cost_power_law_flow(blk, number_of_parallel_units)

        # Register flows - electricity already done by cost_power_law_flow
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.acetic_acid_demand[t0], "acetic_acid"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.phosphoric_acid_demand[t0], "phosphoric_acid"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.ferric_chloride_demand[t0], "ferric_chloride"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.activated_carbon_demand[t0], "activated_carbon"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.sand_demand[t0], "sand"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.anthracite_demand[t0], "anthracite"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.cationic_polymer_demand[t0], "cationic_polymer"
        )



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Dewatering unit model for BSM2 and plant-wide wastewater treatment modeling. 
This unit inherits from the IDAES separator unit.

Model based on

J. Alex, L. Benedetti, J.B. Copp, K.V. Gernaey, U. Jeppsson,
I. Nopens, M.N. Pons, C. Rosen, J.P. Steyer and
P. A. Vanrolleghem
Benchmark Simulation Model no. 2 (BSM2)

Modifications made to TSS formulation based on ASM type.
"""
from enum import Enum, auto

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
)
from idaes.models.unit_models.separator import SeparatorData, SplittingType

from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog

from pyomo.environ import (
    Param,
    units as pyunits,
    Var,
    NonNegativeReals,
)
from pyomo.common.config import ConfigValue, In

from idaes.core.util.exceptions import (
    ConfigurationError,
)
from watertap.costing.unit_models.dewatering import cost_dewatering

__author__ = "Alejandro Garciadiego, Adam Atia"


# Set up logger
_log = idaeslog.getLogger(__name__)


class ActivatedSludgeModelType(Enum):
    """
    ASM1: ASM1 model
    ASM2D: ASM2D model
    modified_ASM2D: modified ASM2D model for ADM1 compatibility
    """

    ASM1 = auto()
    ASM2D = auto()
    modified_ASM2D = auto()


@declare_process_block_class("DewateringUnit")
class DewateringData(SeparatorData):
    """
    Dewatering unit block for BSM2
    """

    CONFIG = SeparatorData.CONFIG()
    CONFIG.outlet_list = ["underflow", "overflow"]
    CONFIG.split_basis = SplittingType.componentFlow

    CONFIG.declare(
        "activated_sludge_model",
        ConfigValue(
            default=ActivatedSludgeModelType.ASM1,
            domain=In(ActivatedSludgeModelType),
            description="Activated Sludge Model used with unit",
            doc="""
        Options to account for version of activated sludge model property package.

        **default** - ``ActivatedSludgeModelType.ASM1``

    .. csv-table::
        :header: "Configuration Options", "Description"

        "``ActivatedSludgeModelType.ASM1``", "ASM1 model"
        "``ActivatedSludgeModelType.ASM2D``", "ASM2D model"
        "``ActivatedSludgeModelType.modified_ASM2D``", "modified ASM2D model for ADM1 compatibility"
    """,
        ),
    )

    def build(self):
        """
        Begin building model.
        Args:
            None
        Returns:
            None
        """

        # Call UnitModel.build to set up dynamics
        super(DewateringData, self).build()

        if "underflow" and "overflow" not in self.config.outlet_list:
            raise ConfigurationError(
                "{} encountered unrecognised "
                "outlet_list. This should not "
                "occur - please use overflow "
                "and underflow as outlets.".format(self.name)
            )

        self.p_dewat = Param(
            initialize=0.28,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Percentage of suspended solids in the underflow",
        )

        self.TSS_rem = Param(
            initialize=0.98,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Percentage of suspended solids removed",
        )

        self.electricity_consumption = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )

        # 0.026 kWh/m3 average between averages of belt and screw presses & centrifuge in relation to flow capacity
        self.energy_electric_flow_vol_inlet = Param(
            self.flowsheet().time,
            units=pyunits.kWh / (pyunits.m**3),
            initialize=0.026,
            mutable=True,
            doc="Specific electricity intensity of unit",
        )

        @self.Constraint(self.flowsheet().time, doc="Electricity consumption equation")
        def eq_electricity_consumption(blk, t):
            return blk.electricity_consumption[t] == pyunits.convert(
                blk.energy_electric_flow_vol_inlet[t] * blk.inlet.flow_vol[t],
                to_units=pyunits.kW,
            )

        self.hydraulic_retention_time = Var(
            self.flowsheet().time,
            initialize=1800,
            domain=NonNegativeReals,
            units=pyunits.s,
            doc="Hydraulic retention time",
        )
        self.volume = Var(
            self.flowsheet().time,
            initialize=1800,
            domain=NonNegativeReals,
            units=pyunits.m**3,
            doc="Hydraulic retention time",
        )

        @self.Constraint(self.flowsheet().time, doc="Hydraulic retention time equation")
        def eq_hydraulic_retention(blk, t):
            return (
                self.hydraulic_retention_time[t]
                == self.volume[t] / self.inlet.flow_vol[t]
            )

        @self.Expression(self.flowsheet().time, doc="Suspended solid concentration")
        def TSS_in(blk, t):
            if blk.config.activated_sludge_model == ActivatedSludgeModelType.ASM1:
                return 0.75 * (
                    sum(
                        blk.inlet.conc_mass_comp[t, i]
                        for i in blk.config.property_package.tss_component_set
                    )
                )
            elif blk.config.activated_sludge_model == ActivatedSludgeModelType.ASM2D:
                return blk.inlet.conc_mass_comp[
                    t, blk.config.property_package.tss_component_set.first()
                ]
            elif (
                blk.config.activated_sludge_model
                == ActivatedSludgeModelType.modified_ASM2D
            ):
                return blk.mixed_state[t].TSS
            else:
                raise ConfigurationError(
                    "The activated_sludge_model was not specified properly in configuration options."
                )

        @self.Expression(self.flowsheet().time, doc="Dewatering factor")
        def f_dewat(blk, t):
            return blk.p_dewat * (10 / (blk.TSS_in[t]))

        @self.Expression(self.flowsheet().time, doc="Remove factor")
        def f_q_du(blk, t):
            return blk.TSS_rem / (pyunits.kg / pyunits.m**3) / 100 / blk.f_dewat[t]

        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.particulate_component_set,
            doc="particulate fraction",
        )
        def overflow_particulate_fraction(blk, t, i):
            return blk.split_fraction[t, "overflow", i] == 1 - blk.TSS_rem

        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.non_particulate_component_set,
            doc="soluble fraction",
        )
        def non_particulate_components(blk, t, i):
            return blk.split_fraction[t, "overflow", i] == 1 - blk.f_q_du[t]

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        param_dict = {}
        for k in self.split_fraction.keys():
            if k[0] == time_point:
                var_dict[f"Split Fraction [{str(k[1:])}]"] = self.split_fraction[k]
        var_dict["Electricity consumption"] = self.electricity_consumption[time_point]
        param_dict[
            "Specific electricity consumption"
        ] = self.energy_electric_flow_vol_inlet[time_point]
        var_dict["Unit Volume"] = self.volume[time_point]
        var_dict["Hydraulic Retention Time"] = self.hydraulic_retention_time[time_point]
        return {"vars": var_dict, "params": param_dict}

    def _get_stream_table_contents(self, time_point=0):
        outlet_list = self.create_outlet_list()

        io_dict = {}
        if self.config.mixed_state_block is None:
            io_dict["Inlet"] = self.mixed_state
        else:
            io_dict["Inlet"] = self.config.mixed_state_block

        for o in outlet_list:
            io_dict[o] = getattr(self, o + "_state")

        return create_stream_table_dataframe(io_dict, time_point=time_point)

    @property
    def default_costing_method(self):
        return cost_dewatering
